-- SqzRuss.lua v0.2

-- A Squeezebox-to-Russound interface program.  Please see the program
-- documentation in SqzRuss_ReadMe_v02.pdf.

-- Author: John Goodman (email: john@goodmanhq.net)
-- ** All right reserved **

require( "socket" )
url = require( "socket.url" )
require( "bit" )

local file = assert(io.open("/var/log/SqzRuss.log", "w"))

-- ---------------------------------------------------------------------
function MainProgram()	

	-- Main body of the program
	
    PrintLog( "Starting SqzRuss Program" )
	-- initialize parameters and global variables
	initialize()
	
    local SleepDelay = pActiveHeartbeat
	local ActiveMode = true
	local StartSleep = nil

	-- connect to the Russound controller (through VSPE)
	ConnectToRussound()
	
	-- connect to Squeezebox Server
    ConnectToSqueezeCenter()
    
	-- get the initial status of all the Russound zones
	GetZoneStatus()
	
	-- loop continuously
    while true do

		-- Read and process the Russound controller message buffer
        ProcessRUBuffer()

		-- if zones are active, echo the squeezebox player displays
		if next( gActiveZoneTbl ) then

			if gSCConnected then

				-- echo the SB player displays to the Russound keypads
				EchoSBDisplays()

			else
				-- wake up processing again
				PrintLog( "Resuming Active Mode" )
				ConnectToSqueezeCenter()
				SleepDelay = pActiveHeartbeat
				ActiveMode = true
			end
		else
			-- if no zones are active slow down processing

			if ActiveMode then
				StartSleep = os.time()
				ActiveMode = false
			else
				if gSCConnected then
					if os.time() - StartSleep > 15 then
						-- disconnect and power off players when inactive for 15 seconds
						PrintLog( "No Zones Active, Entering Inactive Mode" )
						DisconnectFromSqueezeCenter( pPowerOffSqueezeBox )
						SleepDelay = pInactiveHeartbeat
					end
				end
			end
		end

		-- Sleep for a fraction of a second
		socket.sleep( SleepDelay )
	end

end	
-- ---------------------------------------------------------------------
function initialize()
	
	PrintLog( "Initializing program")

	-- read and process the configuration file
    dofile( "/etc/SqzRuss/SqzRuss.ini" )

	if pIniVersion ~= "v02" then
		error( "Configuration file SqzRuss.ini is not the correct version (v02)", 0 )
	end
	
	-- a few global variables
	gSCConnected = false    -- indicates if connected to Squeezebox Server
    gActiveZoneTbl = {}     -- stores the current source for each active zone
	gPartialRULine = nil    -- used to hold partial Russound buffer lines
	gHoldActive = false     -- indicates if a press-and-hold is active

	-- 0X70 is used as the device id for most controllers, 0x71 is used
	-- for the ACA-E5 controller
	pRUDeviceId = 0x70
	if pACAE5Controller then pRUDeviceId = 0x71 end

	-- initialize the Source table with the connected player ids
	gSourceTbl = {}
	
	if pSource1PlayerId then gSourceTbl[ 0 ] = { PlayerId = pSource1PlayerId } end
	if pSource2PlayerId then gSourceTbl[ 1 ] = { PlayerId = pSource2PlayerId } end
	if pSource3PlayerId then gSourceTbl[ 2 ] = { PlayerId = pSource3PlayerId } end
	if pSource4PlayerId then gSourceTbl[ 3 ] = { PlayerId = pSource4PlayerId } end
	if pSource5PlayerId then gSourceTbl[ 4 ] = { PlayerId = pSource5PlayerId } end
	if pSource6PlayerId then gSourceTbl[ 5 ] = { PlayerId = pSource6PlayerId } end
	if pSource7PlayerId then gSourceTbl[ 6 ] = { PlayerId = pSource7PlayerId } end
	if pSource8PlayerId then gSourceTbl[ 7 ] = { PlayerId = pSource8PlayerId } end
	if pSource9PlayerId then gSourceTbl[ 8 ] = { PlayerId = pSource9PlayerId } end
	if pSource10PlayerId then gSourceTbl[ 9 ] = { PlayerId = pSource10PlayerId } end
	if pSource11PlayerId then gSourceTbl[ 10 ] = { PlayerId = pSource11PlayerId } end
	if pSource12PlayerId then gSourceTbl[ 11 ] = { PlayerId = pSource12PlayerId } end
	
	local SourceEntry
	
	for _, SourceEntry in pairs( gSourceTbl ) do
		SourceEntry.LastDisplayText = ""
		SourceEntry.LastDisplayTime = 0
	end
	
	-- pre-format the channel, title, artist, and genre display commands
	gRUDispCmd = string.char( 0x7D, 0x00, 0x79 ) -- target device
	gRUDispCmd = gRUDispCmd .. string.char( 0x00, pRUDeviceId ) -- source device
	gRUDispCmd = gRUDispCmd .. "ss"  -- source no, inserted later
	gRUDispCmd = gRUDispCmd .. string.char( 0x00, 0x02, 0x01, 0x01, 0x02, 0x01, 0x01 )
	gRUDispCmd = gRUDispCmd .. string.char( 0x00, 0x00 ) -- packet no
	gRUDispCmd = gRUDispCmd .. string.char( 0x01, 0x00 ) -- no of packets
	gRUDispCmd = gRUDispCmd .. string.char( pRUDisplayLength + 4, 0x00 ) -- payload length

	-- This table has the hex strings for the different display type
	-- messages.  The "channel" type is used for the main UNO-S2 display.
	-- Other types used to fill-in fields on other keypad types?
	gRUDisplayTypeChars = { channel   = string.char( 0x07, 0x00 ),
							channelno = string.char( 0x06, 0x00 ),
	                        title     = string.char( 0x01, 0x00 ),
                            artist    = string.char( 0x02, 0x00 ),
                            genre     = string.char( 0x03, 0x00 ) }

	pKeyPressTbl = { }
	pKeyPressTbl[ string.char( 0x67, 0x00 ) ] = pPreviousPressCmd   -- Previous
	pKeyPressTbl[ string.char( 0x68, 0x00 ) ] = pNextPressCmd       -- Next
	pKeyPressTbl[ string.char( 0x69, 0x00 ) ] = pPlusPressCmd       -- Plus
	pKeyPressTbl[ string.char( 0x6A, 0x00 ) ] = pMinusPressCmd      -- Minus
    pKeyPressTbl[ string.char( 0x6D, 0x00 ) ] = pStopPressCmd       -- Stop
	pKeyPressTbl[ string.char( 0x6E, 0x00 ) ] = pPausePressCmd      -- Pause
	pKeyPressTbl[ string.char( 0x73, 0x00 ) ] = pPlayPressCmd       -- Play

	pKeyHoldTbl = {}
	pKeyHoldTbl[ string.char( 0xF1, 0x66 ) ] = { Cmd = pPreviousHoldCmd,  -- Previous
											 Rpt = pPreviousHoldRepeat }
	pKeyHoldTbl[ string.char( 0xF1, 0x65 ) ] = { Cmd = pNextHoldCmd,  -- Next
											 Rpt = pNextHoldRepeat }
	pKeyHoldTbl[ string.char( 0xF1, 0x64 ) ] = { Cmd = pPlusHoldCmd,  -- Plus
											 Rpt = pPlusHoldRepeat }
	pKeyHoldTbl[ string.char( 0xF1, 0x63 ) ] = { Cmd = pMinusHoldCmd,  -- Minus
											 Rpt = pMinusHoldRepeat }
	pKeyHoldTbl[ string.char( 0xF1, 0x60 ) ] = { Cmd = pStopHoldCmd,  -- Stop
											 Rpt = pStopHoldRepeat }
	pKeyHoldTbl[ string.char( 0xF1, 0x5F ) ] = { Cmd = pPauseHoldCmd,  -- Pause
											 Rpt = pPauseHoldRepeat }
	pKeyHoldTbl[ string.char( 0xF1, 0x5A ) ] = { Cmd = pPlayHoldCmd,  -- Play
											 Rpt = pPlayHoldRepeat }

	-- Russound command string to request zone status.  "cc" and "zz" are
	-- replaced with the possible combinations of controllers and zones
	-- in the GetZoneStatus() function below
    gZoneStatusRequest = "cc\000\127cczz" .. string.char( pRUDeviceId ) .. "\001\004\002\000zz\007\000\000"

	-- "device reset" message used to initialize a source for an ACA-E5 controller
	gDeviceResetMsg = "\000\000\127\000\000\113\005\002\002\000\000\085\001\003\000\053\000\001"

	return
end
-- ---------------------------------------------------------------------
function FormatHexStr( HexStr )

    -- Returns a print-friendly version of a binary string
	
    local OutStr = ""
    local idx, b
	for idx = 1, #HexStr do
		b = HexStr:sub( idx, idx )
        OutStr = OutStr .. string.format( "%02X:", string.byte( b ))
    end
    return( OutStr )
end
-- ---------------------------------------------------------------------
function ConnectToSqueezeCenter()

	local SocketResult, SocketErr
	local SourceNo, SourceStatus
	local SCResponse, idx

	-- assume not connected
	gSCConnected = false
	
	-- gTcpEnv = socket.tcp
	gTcpSocket, SocketErr = socket.tcp()

	if gTcpSocket then

		SocketResult, SocketErr = gTcpSocket:connect( pSCServerAddr, pSCPortNo )

		if SocketResult then
			gSCConnected = true
		end
	end

    -- td: handle SC login id/password?

	if gSCConnected then
		PrintLog( "Connected to Squeezebox Server at address " ..  pSCServerAddr .. ", port no " .. pSCPortNo )

		-- check for connected players and power-on any squeezeboxes that 
		-- should be active
		for SourceNo, SourceStatus in pairs( gSourceTbl ) do
			-- check for connected players
			SCResponse = SCSendReceive( SourceStatus.PlayerId .. " name ?" )
			idx = SCResponse:find( "name" )
			SCResponse = url.unescape( SCResponse:sub( idx + 5, -1 ) )
			if SCResponse ~= "?" then
				SCResponse = SCResponse .. " (" .. SourceStatus.PlayerId .. ")"
				PrintLog( "Squeezebox Player " .. SCResponse .. " connected as source " .. SourceNo + 1 )
				if isSourceActive( SourceNo ) then
					-- power on the player if it should be active
					SCSendReceive( SourceStatus.PlayerId .. " power 1" )
				end
			else
				PrintLog( "Squeezebox Player Id " .. SourceStatus.PlayerId .. " is not valid/connected" )
			end
		end
	else
		error( "Could not connect to Squeezebox Server at address " ..  pSCServerAddr .. ", port no " .. pSCPortNo, 0 )
	end

	return
end
-- ---------------------------------------------------------------------
function DisconnectFromSqueezeCenter( TurnSqueezeBoxOff )

	-- Close the connection to SqueezeCenter and optionally turn off
	-- all connected Squeezebox players first

	local SourceNo, SourceStatus
	
	-- if parameter set, power off all Squeezebox players first
	if TurnSqueezeBoxOff then
		for SourceNo, SourceStatus in pairs( gSourceTbl ) do
			SCSendReceive( SourceStatus.PlayerId .. " power 0" )
		end
	end

	-- close the socket connection to SqueezeCenter
	gTcpSocket:close()
	gSCConnected = false

	return
end
-- ---------------------------------------------------------------------
function ConnectToRussound()

	local SocketResult, SocketErr

	gRUConnected = false   -- assume no connection
	comsocket = socket.tcp()  -- create socket connection object

	if comsocket then
	
		-- connect to VSPE Tcp socket
		SocketResult, SocketErr = comsocket:connect( pVSPEServerAddr, pVSPEPortNo )
		if SocketResult then
		
			gRUConnected = true
			comsocket:settimeout( 0.025 )
			PrintLog( "Connected to Russound controller at address " ..  pVSPEServerAddr .. ", port no ".. pVSPEPortNo )
			
			if pACAE5Controller then
				-- if an ACAC5 controller, send a device reset message
				SendRUCommand( gDeviceResetMsg )
			end
		end
	end
	
	if not gRUConnected then
		error( "Could not connect to Russound controller/VSPE at address " ..  pVSPEServerAddr .. ", port no " .. pVSPEPortNo, 0 )
	end
	return
	
end
-- ---------------------------------------------------------------------
function SCSendReceive( ScCommand )

	-- Send a CLI command to SqueezeCenter and return the reply

	local SocketResult, SocketErr

	SocketResult, SocketErr = gTcpSocket:send( ScCommand .. "\n" )
	if not SocketResult then
		error("Network error sending Squeezebox Server command, error: " .. SocketErr, 0 )
	else

		SocketResult, SocketErr = gTcpSocket:receive()
		-- print("receive: " .. FormatHexStr( SocketResult ) )
        if not SocketResult then
			error("Network error receiving Squeezebox Server result, error: " .. SocketErr, 0 )
		end
	end
	
	return( SocketResult )
end
-- ---------------------------------------------------------------------
function GetSBDisplay( SourceNo )

	-- This function queries what is currently displayed on the Squeezebox
	-- and returns the result.  This function is called every fraction of
	-- a second to keep the keypad display in-sync with the Squeezebox
	-- display.

	local OutStr = "SB: Not Connected"
	local response, idx, Char
	local SBPlayerId
	
	-- get the status of the current source
	SBPlayerId = gSourceTbl[ SourceNo ][ "PlayerId" ]
	
	-- CLI command to query what's currently displayed on the Squeezebox
    response = SCSendReceive( SBPlayerId .. " display ? ?" )

	if response then
		
		-- skip to the 2nd display line returned
		idx = response:find( "%s", 38 )
		if idx then

			-- unescape the returned control characters
			response = url.unescape( response:sub( idx + 1, -1 ) )

			if response == "?" then
				OutStr = "Player not connected"
			else
				-- remove any unprintable characters (esp. search letters)
				-- that the keypad can't display
				OutStr = ""
				for idx = 1, #response do
					Char = response:byte( idx )
					if Char >= 32 and Char <= 126 then
						OutStr = OutStr .. string.char( Char )
					end
				end
			end
		else
			OutStr = "Wait..."
		end
	end

	return( OutStr )
end
-- ---------------------------------------------------------------------
function EchoSBDisplays()
	--[[
	This function loops through each connected SB player source and
	retrieves what is currently displayed on the player's display.  The
	Russound keypad displays are then updated if the display text has
	changed.  The Russound keypad displays are updated every 30 seconds,
	even if the text hasn't changed, to force the keypad to keep scrolling
	the display text.
	]]

	local SourceNo, SourceStatus
	local CurrDisplayText, PrevDisplayText
	local CurrTime

	-- loop through each SB player source to check for display updates
	for SourceNo, SourceStatus in pairs( gSourceTbl ) do

		if isSourceActive( SourceNo ) then
			CurrTime = os.time()

			-- get the current SB display text
			CurrDisplayText = GetSBDisplay( SourceNo )
			PrevDisplayText = SourceStatus.LastDisplayText

			-- if the display has changed, update the keypad display
			if CurrDisplayText ~= PrevDisplayText then
				SendRUDisplayCommand( SourceNo, "channel", CurrDisplayText )
				SourceStatus.LastDisplayText = CurrDisplayText
				SourceStatus.LastDisplayTime = CurrTime
				PrintLog( "Source " .. SourceNo + 1 .. " Display: " .. CurrDisplayText )
			else
				if CurrDisplayText == "Wait..." then
					if CurrTime - SourceStatus.LastDisplayTime > 5 then
						SendRUDisplayCommand( SourceNo, "channel", "Player Off, Press Play" )
						SourceStatus.LastDisplayTime = CurrTime
					end
				else
					-- force a display update every 30 seconds to maintain scrolling
					if CurrTime - SourceStatus.LastDisplayTime > 30 then
						SendRUDisplayCommand( SourceNo, "channel", CurrDisplayText )
						SourceStatus.LastDisplayTime = CurrTime
					end
				end
			end
		end
	end
	return
end
-- ---------------------------------------------------------------------
function FormatRUCommand( RUCmdStr )

	--[[ This function formats a raw Russound command string.  First, any
	characters with the high-bit set are inverted and pre-pended with
	a 0xF1.  0xF0 is then added to the beginning of the string.
	Finally, the checksum is computed and added to the end of the string
	with a 0xF7.
	]]

	local OutStr = ""
	local idx, CurrChar
	local Mask = 0xFFFFFF00
	local CheckSum = 0

	-- invert any characters with high bit set, prepend 0xF1
	for idx = 1, #RUCmdStr do
		CurrChar = RUCmdStr:byte( idx )
		if CurrChar > 127 then
			OutStr = OutStr .. string.char( 0xF1, bit.bnot( bit.bor( Mask, CurrChar ) ) )
		else
			OutStr = OutStr .. string.char( CurrChar )
		end
	end

	-- Add 0xF0 to the beginning of the command
	OutStr = "\240" .. OutStr

	-- Compute the checksum (sum plus count of the characters)
	for idx = 1, #OutStr do

		CheckSum = CheckSum + OutStr:byte( idx ) + 1

	end

    -- Take only the low bits of the checksum
	CheckSum = bit.band( CheckSum, 0x7F )

	-- add the checksum character and 0xF7 to the end of the command
	OutStr = OutStr .. string.char( CheckSum, 0xF7 )

	return( OutStr )
end
-- ---------------------------------------------------------------------
function SendRUCommand( RUCommand )

	--[[ This function sends a raw command string to the Russound controller.
	The function first calls FormatRUCommand() to format the command string 
	properly, then sends the formatted command to the Russound controller
	through the socket connection (VSPE).  A delay is added after the command 
	is sent to keep the controller from getting overloaded with messages.
	]]
	local SocketResult, SocketErr

	-- format the raw command string (add checksum, etc.)
	RUCommand = FormatRUCommand( RUCommand )

	SocketResult, SocketErr = comsocket:send( RUCommand )
	if not SocketResult then
		error("Network error sending Russound command, error: " .. SocketErr, 0 )
	end

	-- give the controller a bit of a rest between commands
    socket.sleep( pRussoundCommandDelay )  
		
	return
end
-- ---------------------------------------------------------------------
function SendRUDisplayCommand( SourceNo, DisplayType, DisplayText )

	--[[ This function formats and sends a keypad display command to the 
	Russound controller.  
	
	Note that display commands "originate" from a source and all zone
	keypads with the source selected automatically receive the display
	update; i.e., display commands do not need to be sent to each keypad
	separately.
	]]

	local OutStr, NullPad

	if DisplayText ~= nil and DisplayText ~= "" then
		-- start with the base string formatted above
		OutStr = gRUDispCmd
        
		-- insert the source no into the string
        OutStr = OutStr:gsub( "ss", string.char( SourceNo ) )
		
		-- append hex chars for msg type and source
		OutStr = OutStr .. string.char( 32 + SourceNo )
	
		-- append the hex chars for the display type
		OutStr = OutStr .. gRUDisplayTypeChars[ DisplayType ]

		-- display text should be no longer than pRUDisplayLength
		-- characters and should be padded with nulls
		DisplayText = DisplayText:sub( 1, pRUDisplayLength )
		NullPad = string.rep( "\000", pRUDisplayLength + 1 - #DisplayText  )

		-- send the display message to the controller
		SendRUCommand( OutStr .. DisplayText .. NullPad )
	end

	return
end
-- ---------------------------------------------------------------------
function ProcessRUBuffer()

	--[[ This function processes the buffer of Russound messages.  The
	buffer is split into separate lines to be processed, with each line
	ending with 0xF7.  This would normally be pretty easy to do, but the
	last line of the buffer may be a partial line (split because of the
	timeout).  In this case, the partial line needs to be prepended to
	the beginning of the first line the next time a buffer is processed.
	]]

	local buff
	buff = GetRUBuffer()   -- get the Russound buffer

    local lastend = 1
    local findpattern =  "(.-)" .. "\247"   -- lines ending with 0xF7
	local startpos, endpos, line = buff:find( findpattern, 1 )

	while startpos do
		if startpos ~= 1 or line ~= "" then
			if gPartialRULine then
			    -- a partial line was found last time, prepend it to the
				-- beginning of this line
				line = gPartialRULine .. line
				gPartialRULine = nil
			end
            line = line .. buff:sub( endpos, endpos ) -- add the 0xF7 back
			ProcessRUMessage( line )  -- process the message
		end
		lastend = endpos + 1
		startpos, endpos, line = buff:find( findpattern, lastend )
	end

	-- determine if there's a partial line remaining and save it
	if lastend <= #buff then
		gPartialRULine = buff:sub( lastend )
	end

	return
end
-- ---------------------------------------------------------------------
function GetRUBuffer()

	--[[ This function calls the tcp socket connected to the serial
	access to the Russound controller to pull the buffer of queued
	Russound messages.  The partial socket result is used because the
	buffer is received only when the socket timeout occurs.
	]]

	local buffer, err, partial

	buffer, err, partial = comsocket:receive("*all")

	if err ~= "timeout" then
	    error( "Network error reading Russound message buffer, error: " .. err, 0 )
	end

	return( partial )

end
-- ---------------------------------------------------------------------
function ProcessRUMessage( line )

	--[[ This function processes individual messages lines coming from
	the Russound controller.  The individual fields in the message are
	inspected to determine the type of message (zone on/off, source 
	change, key press, etc.) and whether or not this is a message we need
	to process or ignore.  
	]]

	local idx, RuKeyCode, MessageType
	local zone, source, SourceStatus

	-- print( FormatHexStr( line ) )

    MessageType = line:sub( 8, 8 )

	-- if an event type message
	if MessageType == "\005" then

		-- skip past target and source paths
		idx = 10 + line:byte( 9 )
		idx = idx + line:byte( idx ) + 1

		-- Hex code for the event type or key pressed
		RuKeyCode = line:sub( idx, idx + 1 )

		-- if a zone on/off event message
		if RuKeyCode == "\241\035" then
			zone = line:sub( 2, 3 )
			-- 0x00 for off, 0x01 for on
			if line:sub( idx + 3, idx + 3 ) == "\000" then
				PrintLog( "Zone " .. FormatHexStr( zone ) .. " turned off" )
				ProcessZoneChange( zone, nil )
			else
				-- PrintLog( "Event: Zone On " .. FormatHexStr( zone ) )
			end
		else  -- a keypress message
			-- pull out the source zone
			zone = line:sub( 5, 6 )

			-- Process the Keypress
			ProcessKeyPress( zone, RuKeyCode )
		end

	-- if a source name display message, indicating the zone is changing sources
	-- (msg is also sent when a zone is turned on)
	elseif MessageType == "\006" and line:sub( 13, 13 ) == "\005" then
        zone = line:sub( 2, 3 )
        source = line:byte( 10 )
        ProcessZoneChange( zone, source )
		
	-- Response to a "Get All Zone Info" request.  Byte 21 is the zone on/off state
	-- and byte 22 is the zone's currently selected source
	elseif MessageType == "\000" then
        zone = line:sub( 2, 3 )
		source = line:byte( 22 )
        if line:sub( 21, 21 ) == "\001" then
            ProcessZoneChange( zone, source )
        end
    end

	return
end
-- ---------------------------------------------------------------------
function ProcessKeyPress( ZoneNo, KeyCode )

	--[[ This function processes a key press message from the Russound
	controller.  We first determine if the key press is from a zone that
	has a Squeezebox source and if it's a key we're interested in 
	processing.  If so, we then look up and send the corresponding 
	command to the Squeezebox player.  We also need to keep track of press
	and press-and-hold state to process the key press correctly.  This
	logic doesn't properly handle two keypads doing a press-and-hold at
	the same time, but it should be good enough for most uses.
	]]

	local SourceNo, SbKey, SbKeyTbl

	SourceNo = gActiveZoneTbl[ ZoneNo ]

	-- if the keypress is from a zone with a SqueezeBox as the selected source
	if SourceNo then

		-- if a keypress for a key we're interested in
		SbKey = pKeyPressTbl[ KeyCode ]
		if SbKey then
			if gHoldActive then
				gHoldActive = false
			else
				PrintLog( "Keypress Zone: " .. FormatHexStr( ZoneNo ) .. SbKey )
				SCSendReceive( gSourceTbl[ SourceNo ][ "PlayerId" ] .. " " .. SbKey )
			end
		else
			-- if a press-and-hold for a key we're interested in
			SbKeyTbl = pKeyHoldTbl[ KeyCode ]

			if SbKeyTbl then
				-- send the key command to the squeezebox if it's the
				-- first time the press-and-hold detected or if
				-- we repeat the key on every occurence
				if not gHoldActive or SbKeyTbl[ "Rpt" ] then
					SbKey = SbKeyTbl[ "Cmd" ]
					PrintLog( "Keypress Zone: " .. FormatHexStr( ZoneNo ) .. SbKey )
					SCSendReceive( gSourceTbl[ SourceNo ][ "PlayerId" ] .. " " .. SbKey )
				end
				-- mark that a press-and-hold is active
				gHoldActive = true
			end
		end
	end
	return
end
-- ---------------------------------------------------------------------
function ProcessZoneChange( ZoneNo, SourceNo )

	--[[ This function processes a zone on/off or zone source change
	message.  
	]]

    -- Assume the zone is no longer active
	gActiveZoneTbl[ ZoneNo ] = nil

	if SourceNo then
		
		PrintLog( "Zone " .. FormatHexStr( ZoneNo ) .. " source set to " .. SourceNo + 1 )

		-- if this is a Squeezebox source, mark the zone as active
		if gSourceTbl[ SourceNo ] then
			gActiveZoneTbl[ ZoneNo ] = SourceNo
			-- force keypad display update
			gSourceTbl[ SourceNo ][ "LastDisplayText" ] = ""
			-- make sure the player is powered on
			if gSCConnected then
				SCSendReceive( gSourceTbl[ SourceNo ][ "PlayerId" ] .. " power 1" )
			end
			PrintLog( "Active Squeezebox zones/sources:" )
			local k, v
			for k,v in pairs( gActiveZoneTbl ) do PrintLog( "  Zone: " .. FormatHexStr(k) .. " source: " .. v + 1 ) end
		end	
	end
	
	if not next( gActiveZoneTbl ) then
		PrintLog( "No active Squeezebox zones" )
	end
	
	return
end
-- ---------------------------------------------------------------------
function GetZoneStatus()

	-- Send zone status requests to the Russound controller for each
	-- possible combination of controller (0 to 5) and zone (0 to 5).
	-- Note that the replies are asynchronous and are processed
	-- with other controller messages above.

    local Controller, Zone, ControllerChar, ZoneChar, Request

    gActiveZoneTbl = {}

    for Controller = 0, 5 do
        ControllerChar = string.char( Controller )
        for Zone = 0, 5 do
            ZoneChar = string.char( Zone )
			-- substitute the controller and zone hex chars in the base msg
            Request = gZoneStatusRequest:gsub( "cc", ControllerChar )
            Request = Request:gsub( "zz", ZoneChar )
            -- print( FormatHexStr( Request ) )
            SendRUCommand( Request )
        end
    end
    return
end
-- ---------------------------------------------------------------------
function isSourceActive( SourceNo )

	-- Returns true if the passed source number is currently active in
	-- any zone
	
	local source
	
	for _, source in pairs( gActiveZoneTbl ) do
		if source == SourceNo then return( true ) end
	end

	return( false )
end
-- ---------------------------------------------------------------------
function PrintLog( LogStr )
	file:write(os.date().." - "..LogStr.."\n")
end
-- ---------------------------------------------------------------------

	-- call the main program function, trapping any fatal errors
	-- during execution
	
	local PgmStatus, ErrorMsg = pcall( MainProgram )
	if not PgmStatus then

		PrintLog( "--------------------------------------------" )
		PrintLog( "The SqzRuss program ended with a fatal error" )
		PrintLog( " " )
		PrintLog( ErrorMsg )
		PrintLog( " " )

	end

	file:close()	

	